@@ -32,6 +32,7 @@ class CsCompiler(Compiler):
         self.language = 'cs'
         super().__init__(exelist, version)
         self.id = id
+        self.is_cross = False
         self.runner = runner
 
     def get_display_language(self):
@@ -88,6 +89,15 @@ class CsCompiler(Compiler):
     def get_pic_args(self):
         return []
 
+    def compute_parameters_with_absolute_paths(self, parameter_list, build_dir):
+        for idx, i in enumerate(parameter_list):
+            if i[:2] == '-L':
+                parameter_list[idx] = i[:2] + os.path.normpath(os.path.join(build_dir, i[2:]))
+            if i[:5] == '-lib:':
+                parameter_list[idx] = i[:5] + os.path.normpath(os.path.join(build_dir, i[5:]))
+
+        return parameter_list
+
     def name_string(self):
         return ' '.join(self.exelist)
 
