@@ -11,16 +11,28 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
+from typing import List
+import subprocess, os
 
 from .c import CCompiler
 from .compilers import (
-    ICC_STANDARD,
+    CompilerType,
+    apple_buildtype_linker_args,
+    gnulike_buildtype_args,
+    gnulike_buildtype_linker_args,
+    gnu_optimization_args,
+    clike_debug_args,
     Compiler,
     GnuCompiler,
+    ClangCompiler,
     ElbrusCompiler,
     IntelCompiler,
+    PGICompiler
 )
 
+from mesonbuild.mesonlib import EnvironmentException, MachineChoice, is_osx
+
+
 class FortranCompiler(Compiler):
     library_dirs_cache = CCompiler.library_dirs_cache
     program_dirs_cache = CCompiler.library_dirs_cache
@@ -61,6 +73,48 @@ class FortranCompiler(Compiler):
     def get_soname_args(self, *args):
         return CCompiler.get_soname_args(self, *args)
 
+    def sanity_check(self, work_dir, environment):
+        source_name = os.path.join(work_dir, 'sanitycheckf.f90')
+        binary_name = os.path.join(work_dir, 'sanitycheckf')
+        with open(source_name, 'w') as ofile:
+            ofile.write('print *, "Fortran compilation is working."; end')
+        extra_flags = []
+        if self.is_cross:
+            extra_flags = environment.coredata.get_external_args(MachineChoice.HOST, self.language)
+            extra_flags += environment.coredata.get_external_link_args(MachineChoice.HOST, self.language)
+        pc = subprocess.Popen(self.exelist + extra_flags + [source_name, '-o', binary_name])
+        pc.wait()
+        if pc.returncode != 0:
+            raise EnvironmentException('Compiler %s can not compile programs.' % self.name_string())
+        if self.is_cross:
+            if self.exe_wrapper is None:
+                # Can't check if the binaries run so we have to assume they do
+                return
+            cmdlist = self.exe_wrapper + [binary_name]
+        else:
+            cmdlist = [binary_name]
+        pe = subprocess.Popen(cmdlist, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
+        pe.wait()
+        if pe.returncode != 0:
+            raise EnvironmentException('Executables created by Fortran compiler %s are not runnable.' % self.name_string())
+
+    def get_std_warn_args(self, level):
+        return FortranCompiler.std_warn_args
+
+    def get_buildtype_args(self, buildtype):
+        return gnulike_buildtype_args[buildtype]
+
+    def get_optimization_args(self, optimization_level):
+        return gnu_optimization_args[optimization_level]
+
+    def get_debug_args(self, is_debug):
+        return clike_debug_args[is_debug]
+
+    def get_buildtype_linker_args(self, buildtype):
+        if is_osx():
+            return apple_buildtype_linker_args[buildtype]
+        return gnulike_buildtype_linker_args[buildtype]
+
     def split_shlib_to_parts(self, fname):
         return CCompiler.split_shlib_to_parts(self, fname)
 
@@ -119,19 +173,29 @@ class FortranCompiler(Compiler):
         return ('-I', )
 
     def get_module_outdir_args(self, path):
-        return ['-module' + path]
+        return ['-module', path]
 
-    def module_name_to_filename(self, module_name):
+    def compute_parameters_with_absolute_paths(self, parameter_list, build_dir):
+        for idx, i in enumerate(parameter_list):
+            if i[:2] == '-I' or i[:2] == '-L':
+                parameter_list[idx] = i[:2] + os.path.normpath(os.path.join(build_dir, i[2:]))
+
+        return parameter_list
+
+    def module_name_to_filename(self, module_name: str) -> str:
         return module_name.lower() + '.mod'
 
     def get_std_shared_lib_link_args(self):
         return CCompiler.get_std_shared_lib_link_args(self)
 
-    def get_library_dirs_real(self):
-        return CCompiler.get_library_dirs_real(self)
+    def _get_search_dirs(self, *args, **kwargs):
+        return CCompiler._get_search_dirs(self, *args, **kwargs)
 
-    def get_library_dirs(self):
-        return CCompiler.get_library_dirs(self)
+    def get_compiler_dirs(self, *args, **kwargs):
+        return CCompiler.get_compiler_dirs(self, *args, **kwargs)
+
+    def get_library_dirs(self, *args, **kwargs):
+        return CCompiler.get_library_dirs(self, *args, **kwargs)
 
     def get_pic_args(self):
         return CCompiler.get_pic_args(self)
@@ -151,27 +215,22 @@ class FortranCompiler(Compiler):
     def gen_import_library_args(self, implibname):
         return CCompiler.gen_import_library_args(self, implibname)
 
-    def sanity_check(self, work_dir, environment):
-        code = '''program main
-            integer :: ret = 0
-            call exit(ret)
-        end program main'''
-        return CCompiler.sanity_check_impl(self, work_dir, environment, 'sanitycheckf.f90', code)
-
     def _get_compiler_check_args(self, env, extra_args, dependencies, mode='compile'):
         return CCompiler._get_compiler_check_args(self, env, extra_args, dependencies, mode='compile')
 
-    def compiles(self, code, env, extra_args=None, dependencies=None, mode='compile'):
-        return CCompiler.compiles(self, code, env, extra_args, dependencies, mode)
+    def compiles(self, code, env, *, extra_args=None, dependencies=None, mode='compile'):
+        return CCompiler.compiles(self, code, env, extra_args=extra_args,
+                                  dependencies=dependencies, mode=mode)
 
     def _build_wrapper(self, code, env, extra_args, dependencies=None, mode='compile', want_output=False):
         return CCompiler._build_wrapper(self, code, env, extra_args, dependencies, mode, want_output)
 
-    def links(self, code, env, extra_args=None, dependencies=None):
-        return CCompiler.links(self, code, env, extra_args, dependencies)
+    def links(self, code, env, *, extra_args=None, dependencies=None):
+        return CCompiler.links(self, code, env, extra_args=extra_args,
+                               dependencies=dependencies)
 
-    def run(self, code, env, extra_args=None, dependencies=None):
-        return CCompiler.run(self, code, env, extra_args, dependencies)
+    def run(self, code, env, *, extra_args=None, dependencies=None):
+        return CCompiler.run(self, code, env, extra_args=extra_args, dependencies=dependencies)
 
     def _get_patterns(self, *args, **kwargs):
         return CCompiler._get_patterns(self, *args, **kwargs)
@@ -206,13 +265,27 @@ class FortranCompiler(Compiler):
     def has_multi_arguments(self, args, env):
         return CCompiler.has_multi_arguments(self, args, env)
 
+    def has_header(self, hname, prefix, env, *, extra_args=None, dependencies=None):
+        return CCompiler.has_header(self, hname, prefix, env, extra_args=extra_args, dependencies=dependencies)
+
+    def get_define(self, dname, prefix, env, extra_args, dependencies):
+        return CCompiler.get_define(self, dname, prefix, env, extra_args, dependencies)
+
+    @classmethod
+    def _get_trials_from_pattern(cls, pattern, directory, libname):
+        return CCompiler._get_trials_from_pattern(pattern, directory, libname)
+
+    @staticmethod
+    def _get_file_from_list(env, files: List[str]) -> str:
+        return CCompiler._get_file_from_list(env, files)
 
 class GnuFortranCompiler(GnuCompiler, FortranCompiler):
-    def __init__(self, exelist, version, gcc_type, is_cross, exe_wrapper=None, defines=None, **kwargs):
+    def __init__(self, exelist, version, compiler_type, is_cross, exe_wrapper=None, defines=None, **kwargs):
         FortranCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwargs)
-        GnuCompiler.__init__(self, gcc_type, defines)
+        GnuCompiler.__init__(self, compiler_type, defines)
         default_warn_args = ['-Wall']
-        self.warn_args = {'1': default_warn_args,
+        self.warn_args = {'0': [],
+                          '1': default_warn_args,
                           '2': default_warn_args + ['-Wextra'],
                           '3': default_warn_args + ['-Wextra', '-Wpedantic']}
 
@@ -226,20 +299,21 @@ class GnuFortranCompiler(GnuCompiler, FortranCompiler):
         return ['-J' + path]
 
     def language_stdlib_only_link_flags(self):
-        return ['-lgfortran', '-lm', '-lquadmath']
+        return ['-lgfortran', '-lm']
 
 
 class ElbrusFortranCompiler(GnuFortranCompiler, ElbrusCompiler):
-    def __init__(self, exelist, version, gcc_type, is_cross, exe_wrapper=None, defines=None, **kwargs):
-        GnuFortranCompiler.__init__(self, exelist, version, gcc_type, is_cross, exe_wrapper, defines, **kwargs)
-        ElbrusCompiler.__init__(self, gcc_type, defines)
+    def __init__(self, exelist, version, compiler_type, is_cross, exe_wrapper=None, defines=None, **kwargs):
+        GnuFortranCompiler.__init__(self, exelist, version, compiler_type, is_cross, exe_wrapper, defines, **kwargs)
+        ElbrusCompiler.__init__(self, compiler_type, defines)
 
 class G95FortranCompiler(FortranCompiler):
     def __init__(self, exelist, version, is_cross, exe_wrapper=None, **kwags):
         FortranCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwags)
         self.id = 'g95'
         default_warn_args = ['-Wall']
-        self.warn_args = {'1': default_warn_args,
+        self.warn_args = {'0': [],
+                          '1': default_warn_args,
                           '2': default_warn_args + ['-Wextra'],
                           '3': default_warn_args + ['-Wextra', '-pedantic']}
 
@@ -281,23 +355,34 @@ class IntelFortranCompiler(IntelCompiler, FortranCompiler):
         FortranCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwags)
         # FIXME: Add support for OS X and Windows in detect_fortran_compiler so
         # we are sent the type of compiler
-        IntelCompiler.__init__(self, ICC_STANDARD)
+        IntelCompiler.__init__(self, CompilerType.ICC_STANDARD)
         self.id = 'intel'
         default_warn_args = ['-warn', 'general', '-warn', 'truncated_source']
-        self.warn_args = {'1': default_warn_args,
+        self.warn_args = {'0': [],
+                          '1': default_warn_args,
                           '2': default_warn_args + ['-warn', 'unused'],
                           '3': ['-warn', 'all']}
 
     def get_preprocess_only_args(self):
         return ['-cpp', '-EP']
 
+    def get_always_args(self):
+        """Ifort doesn't have -pipe."""
+        val = super().get_always_args()
+        val.remove('-pipe')
+        return val
+
+    def language_stdlib_only_link_flags(self):
+        return ['-lifcore', '-limf']
+
 
 class PathScaleFortranCompiler(FortranCompiler):
     def __init__(self, exelist, version, is_cross, exe_wrapper=None, **kwags):
         FortranCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwags)
         self.id = 'pathscale'
         default_warn_args = ['-fullwarn']
-        self.warn_args = {'1': default_warn_args,
+        self.warn_args = {'0': [],
+                          '1': default_warn_args,
                           '2': default_warn_args,
                           '3': default_warn_args}
 
@@ -305,31 +390,30 @@ class PathScaleFortranCompiler(FortranCompiler):
         return ['-mp']
 
 
-class PGIFortranCompiler(FortranCompiler):
+class PGIFortranCompiler(PGICompiler, FortranCompiler):
     def __init__(self, exelist, version, is_cross, exe_wrapper=None, **kwags):
         FortranCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwags)
-        self.id = 'pgi'
+        PGICompiler.__init__(self, CompilerType.PGI_STANDARD)
+
+
+class FlangFortranCompiler(ClangCompiler, FortranCompiler):
+    def __init__(self, exelist, version, is_cross, exe_wrapper=None, **kwags):
+        FortranCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwags)
+        ClangCompiler.__init__(self, CompilerType.CLANG_STANDARD)
+        self.id = 'flang'
         default_warn_args = ['-Minform=inform']
-        self.warn_args = {'1': default_warn_args,
+        self.warn_args = {'0': [],
+                          '1': default_warn_args,
                           '2': default_warn_args,
                           '3': default_warn_args}
 
-    def get_module_incdir_args(self):
-        return ('-module', )
-
-    def get_no_warn_args(self):
-        return ['-silent']
-
-    def openmp_flags(self):
-        return ['-fopenmp']
-
-
 class Open64FortranCompiler(FortranCompiler):
     def __init__(self, exelist, version, is_cross, exe_wrapper=None, **kwags):
         FortranCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwags)
         self.id = 'open64'
         default_warn_args = ['-fullwarn']
-        self.warn_args = {'1': default_warn_args,
+        self.warn_args = {'0': [],
+                          '1': default_warn_args,
                           '2': default_warn_args,
                           '3': default_warn_args}
 
