@@ -15,18 +15,20 @@
 from . import mlog
 import pickle, os, uuid, shlex
 import sys
+from itertools import chain
 from pathlib import PurePath
 from collections import OrderedDict
 from .mesonlib import (
-    MesonException, default_libdir, default_libexecdir, default_prefix
+    MesonException, MachineChoice, PerMachine,
+    default_libdir, default_libexecdir, default_prefix, stringlistify
 )
 from .wrap import WrapMode
 import ast
 import argparse
 import configparser
 
-version = '0.49.1'
-backendlist = ['ninja', 'vs', 'vs2010', 'vs2015', 'vs2017', 'xcode']
+version = '0.50.1'
+backendlist = ['ninja', 'vs', 'vs2010', 'vs2015', 'vs2017', 'vs2019', 'xcode']
 
 default_yielding = False
 
@@ -42,6 +44,9 @@ class UserOption:
             raise MesonException('Value of "yielding" must be a boolean.')
         self.yielding = yielding
 
+    def printable_value(self):
+        return self.value
+
     # Check that the input is a valid value and return the
     # "cleaned" or "native" version. For example the Boolean
     # option could take the string "true" and return True.
@@ -113,9 +118,14 @@ class UserUmaskOption(UserIntegerOption):
         super().__init__(name, description, 0, 0o777, value, yielding)
         self.choices = ['preserve', '0000-0777']
 
+    def printable_value(self):
+        if self.value == 'preserve':
+            return self.value
+        return format(self.value, '04o')
+
     def validate_value(self, value):
         if value is None or value == 'preserve':
-            return None
+            return 'preserve'
         return super().validate_value(value)
 
     def toint(self, valuestring):
@@ -225,47 +235,11 @@ def load_configs(filenames):
 
             raise MesonException('Cannot find specified native file: ' + f)
 
-    config = configparser.SafeConfigParser()
+    config = configparser.ConfigParser()
     config.read(gen())
     return config
 
 
-def _get_section(config, section):
-    if config.has_section(section):
-        final = {}
-        for k, v in config.items(section):
-            # Windows paths...
-            v = v.replace('\\', '\\\\')
-            try:
-                final[k] = ast.literal_eval(v)
-            except SyntaxError:
-                raise MesonException(
-                    'Malformed value in native file variable: {}'.format(v))
-        return final
-    return {}
-
-
-class ConfigData:
-
-    """Contains configuration information provided by the user for the build."""
-
-    def __init__(self, config=None):
-        if config:
-            self.binaries = _get_section(config, 'binaries')
-            # global is a keyword and globals is a builtin, rather than mangle it,
-            # use a similar word
-            self.universal = _get_section(config, 'globals')
-            self.subprojects = {s: _get_section(config, s) for s in config.sections()
-                                if s not in {'binaries', 'globals'}}
-        else:
-            self.binaries = {}
-            self.universal = {}
-            self.subprojects = {}
-
-    def get_binaries(self, name):
-        return self.binaries.get(name, None)
-
-
 # This class contains all data that must persist over multiple
 # invocations of Meson. It is roughly the same thing as
 # cmakecache.
@@ -288,9 +262,9 @@ class CoreData:
         self.init_builtins()
         self.backend_options = {}
         self.user_options = {}
-        self.compiler_options = {}
+        self.compiler_options = PerMachine({}, {}, {})
         self.base_options = {}
-        self.external_preprocess_args = {} # CPPFLAGS only
+        self.external_preprocess_args = PerMachine({}, {}, {}) # CPPFLAGS only
         self.cross_file = self.__load_cross_file(options.cross_file)
         self.compilers = OrderedDict()
         self.cross_compilers = OrderedDict()
@@ -298,6 +272,7 @@ class CoreData:
         # Only to print a warning if it changes between Meson invocations.
         self.pkgconf_envvar = os.environ.get('PKG_CONFIG_PATH', '')
         self.config_files = self.__load_config_files(options.native_file)
+        self.libdir_cross_fixup()
 
     @staticmethod
     def __load_config_files(filenames):
@@ -347,6 +322,13 @@ class CoreData:
 
         raise MesonException('Cannot find specified cross file: ' + filename)
 
+    def libdir_cross_fixup(self):
+        # By default set libdir to "lib" when cross compiling since
+        # getting the "system default" is always wrong on multiarch
+        # platforms as it gets a value like lib/x86_64-linux-gnu.
+        if self.cross_file is not None:
+            self.builtins['libdir'].value = 'lib'
+
     def sanitize_prefix(self, prefix):
         if not os.path.isabs(prefix):
             raise MesonException('prefix value {!r} must be an absolute path'
@@ -476,21 +458,34 @@ class CoreData:
             mode = 'custom'
         self.builtins['buildtype'].set_value(mode)
 
+    def get_all_compiler_options(self):
+        # TODO think about cross and command-line interface. (Only .build is mentioned here.)
+        yield self.compiler_options.build
+
+    def _get_all_nonbuiltin_options(self):
+        yield self.backend_options
+        yield self.user_options
+        yield from self.get_all_compiler_options()
+        yield self.base_options
+
+    def get_all_options(self):
+        return chain([self.builtins], self._get_all_nonbuiltin_options())
+
     def validate_option_value(self, option_name, override_value):
-        for opts in (self.builtins, self.base_options, self.compiler_options, self.user_options):
+        for opts in self.get_all_options():
             if option_name in opts:
                 opt = opts[option_name]
                 return opt.validate_value(override_value)
         raise MesonException('Tried to validate unknown option %s.' % option_name)
 
-    def get_external_args(self, lang):
-        return self.compiler_options[lang + '_args'].value
+    def get_external_args(self, for_machine: MachineChoice, lang):
+        return self.compiler_options[for_machine][lang + '_args'].value
 
-    def get_external_link_args(self, lang):
-        return self.compiler_options[lang + '_link_args'].value
+    def get_external_link_args(self, for_machine: MachineChoice, lang):
+        return self.compiler_options[for_machine][lang + '_link_args'].value
 
-    def get_external_preprocess_args(self, lang):
-        return self.external_preprocess_args[lang]
+    def get_external_preprocess_args(self, for_machine: MachineChoice, lang):
+        return self.external_preprocess_args[for_machine][lang]
 
     def merge_user_options(self, options):
         for (name, value) in options.items():
@@ -501,7 +496,7 @@ class CoreData:
                 if type(oldval) != type(value):
                     self.user_options[name] = value
 
-    def set_options(self, options, subproject=''):
+    def set_options(self, options, subproject='', warn_unknown=True):
         # Set prefix first because it's needed to sanitize other options
         prefix = self.builtins['prefix'].value
         if 'prefix' in options:
@@ -517,26 +512,121 @@ class CoreData:
                 pass
             elif k in self.builtins:
                 self.set_builtin_option(k, v)
-            elif k in self.backend_options:
-                tgt = self.backend_options[k]
-                tgt.set_value(v)
-            elif k in self.user_options:
-                tgt = self.user_options[k]
-                tgt.set_value(v)
-            elif k in self.compiler_options:
-                tgt = self.compiler_options[k]
-                tgt.set_value(v)
-            elif k in self.base_options:
-                tgt = self.base_options[k]
-                tgt.set_value(v)
             else:
-                unknown_options.append(k)
-
-        if unknown_options:
+                for opts in self._get_all_nonbuiltin_options():
+                    if k in opts:
+                        tgt = opts[k]
+                        tgt.set_value(v)
+                        break
+                else:
+                    unknown_options.append(k)
+        if unknown_options and warn_unknown:
             unknown_options = ', '.join(sorted(unknown_options))
             sub = 'In subproject {}: '.format(subproject) if subproject else ''
             mlog.warning('{}Unknown options: "{}"'.format(sub, unknown_options))
 
+    def set_default_options(self, default_options, subproject, env):
+        # Set defaults first from conf files (cross or native), then
+        # override them as nec as necessary.
+        for k, v in env.paths.host:
+            if v is not None:
+                env.cmd_line_options.setdefault(k, v)
+
+        # Set default options as if they were passed to the command line.
+        # Subprojects can only define default for user options.
+        from . import optinterpreter
+        for k, v in default_options.items():
+            if subproject:
+                if optinterpreter.is_invalid_name(k):
+                    continue
+                k = subproject + ':' + k
+            env.cmd_line_options.setdefault(k, v)
+
+        # Create a subset of cmd_line_options, keeping only options for this
+        # subproject. Also take builtin options if it's the main project.
+        # Language and backend specific options will be set later when adding
+        # languages and setting the backend (builtin options must be set first
+        # to know which backend we'll use).
+        options = {}
+        for k, v in env.cmd_line_options.items():
+            if subproject:
+                if not k.startswith(subproject + ':'):
+                    continue
+            elif k not in get_builtin_options():
+                if ':' in k:
+                    continue
+                if optinterpreter.is_invalid_name(k):
+                    continue
+            options[k] = v
+
+        self.set_options(options, subproject)
+
+    def process_new_compilers(self, lang: str, comp, cross_comp, env):
+        from . import compilers
+
+        self.compilers[lang] = comp
+        if cross_comp is not None:
+            self.cross_compilers[lang] = cross_comp
+
+        # Native compiler always exist so always add its options.
+        new_options_for_build = comp.get_and_default_options(env.properties.build)
+        preproc_flags_for_build = comp.get_preproc_flags()
+        if cross_comp is not None:
+            new_options_for_host = cross_comp.get_and_default_options(env.properties.host)
+            preproc_flags_for_host = cross_comp.get_preproc_flags()
+        else:
+            new_options_for_host = new_options_for_build
+            preproc_flags_for_host = comp.get_preproc_flags()
+
+        opts_machines_list = [
+            (new_options_for_build, preproc_flags_for_build, MachineChoice.BUILD),
+            (new_options_for_host, preproc_flags_for_host, MachineChoice.HOST),
+        ]
+
+        optprefix = lang + '_'
+        for new_options, preproc_flags, for_machine in opts_machines_list:
+            for k, o in new_options.items():
+                if not k.startswith(optprefix):
+                    raise MesonException('Internal error, %s has incorrect prefix.' % k)
+                if (env.machines.matches_build_machine(for_machine) and
+                        k in env.cmd_line_options):
+                    # TODO think about cross and command-line interface.
+                    o.set_value(env.cmd_line_options[k])
+                if k in ['c_std', 'cpp_std']:
+                    # These CLI options should affect all platforms
+                    for m in iter(MachineChoice):
+                        self.compiler_options[m].setdefault(k, o)
+                else:
+                    # By default, CLI options just affect the native platform
+                    self.compiler_options[for_machine].setdefault(k, o)
+
+            # Unlike compiler and linker flags, preprocessor flags are not in
+            # compiler_options because they are not visible to user.
+            preproc_flags = shlex.split(preproc_flags)
+            k = lang + '_args'
+            if lang in ('c', 'cpp', 'objc', 'objcpp') and k in env.properties[for_machine]:
+                # `c_args` in the cross file are used, like CPPFLAGS but *not*
+                # CFLAGS, for tests. this is weird, but how it was already
+                # implemented. Hopefully a new version of #3916 fixes it.
+                preproc_flags = stringlistify(env.properties[for_machine][k])
+            self.external_preprocess_args[for_machine].setdefault(lang, preproc_flags)
+
+        enabled_opts = []
+        for optname in comp.base_options:
+            if optname in self.base_options:
+                continue
+            oobj = compilers.base_options[optname]
+            if optname in env.cmd_line_options:
+                oobj.set_value(env.cmd_line_options[optname])
+                enabled_opts.append(optname)
+            self.base_options[optname] = oobj
+        self.emit_base_options_warnings(enabled_opts)
+
+    def emit_base_options_warnings(self, enabled_opts: list):
+        if 'b_bitcode' in enabled_opts:
+            mlog.warning('Base option \'b_bitcode\' is enabled, which is incompatible with many linker options. Incompatible options such as such as \'b_asneeded\' have been disabled.')
+            mlog.warning('Please see https://mesonbuild.com/Builtin-options.html#Notes_about_Apple_Bitcode_support for more details.')
+
 class CmdLineFileParser(configparser.ConfigParser):
     def __init__(self):
         # We don't want ':' as key delimiter, otherwise it would break when
@@ -747,7 +837,7 @@ builtin_options = {
     'localstatedir':   [UserStringOption, 'Localstate data directory', 'var'],
     'sharedstatedir':  [UserStringOption, 'Architecture-independent data directory', 'com'],
     'werror':          [UserBooleanOption, 'Treat warnings as errors', False],
-    'warning_level':   [UserComboOption, 'Compiler warning level to use', ['1', '2', '3'], '1'],
+    'warning_level':   [UserComboOption, 'Compiler warning level to use', ['0', '1', '2', '3'], '1'],
     'layout':          [UserComboOption, 'Build directory layout', ['mirror', 'flat'], 'mirror'],
     'default_library': [UserComboOption, 'Default library type', ['shared', 'static', 'both'], 'shared'],
     'backend':         [UserComboOption, 'Backend to use', backendlist, 'ninja'],
